#! /usr/bin/env ruby
# coding: utf-8

#Site with element info.
#This doesn't have coordinates.
#
class Crysna::Atom
  attr_reader :element
  attr_reader :site

  #
  def initialize(element, sitename, global_vector)
    @element = element
    @site = Crysna::Site.new(sitename, global_vector)
  end

  def ==(other)
    return false unless self.element == other.element
    return false unless self.site == other.site
    return true
  end

  def migrate!(site_name, global_vector)
    new_gvec = @site.global_vector + global_vector.to_v3di
    @site = Crysna::Site.new(site_name, new_gvec)
  end

  def migrate(site_name, global_vector)
    result = Marshal.load(Marshal.dump(self))
    result.migrate!(site_name, global_vector)
    result
  end

  # Return center-migrate atom.
  # Therefore, global_vector becomes [0,0,0]
  def centering
    result = Marshal.load(Marshal.dump(self))
    result.centering!
    result
  end

  # Destructive migrate to center cell.
  # Therefore, global_vector becomes [0,0,0]
  def centering!
    self.migrate!(@site.name, @site.global_vector * (-1))
  end


  #Mainly for sort
  def <=>(other)
    flag = @element <=> other.element
    return flag unless flag == 0

    flag = @site.name <=> other.site.name
    return flag unless flag == 0

    flag = @site.global_vector[0] <=> other.site.global_vector[0]
    return flag unless flag == 0

    flag = @site.global_vector[1] <=> other.site.global_vector[1]
    return flag unless flag == 0

    flag = @site.global_vector[2] <=> other.site.global_vector[2]
    return flag
  end

  ##Not-destructive method for site migration.
  #def migrate(name, g_vector)
  #    result = Marshal.load(Marshal.dump(self))
  #    result.migrate!(name, g_vector)
  #    return result
  #end

  ##Destructive method for site migration.
  #def migrate!(name, g_vector)
  #    @name = name
  #    @global_vector.size.times do |i|
  #        @global_vector[i] += g_vector[i]
  #    end
  #end

  #Return site name. E.g., "B"
  def sitename
    @site.name
  end

  #Return site name with suffix as global vector with the manner of cif.
  #E.g., "B555"
  def full_sitename
    result = sitename
    @site.global_vector.each do |i|
      result += (i+5).to_i.to_s
    end
    result
  end
end

